/*
 * juked - a jukebox daemon
 * Copyright (C) 2000-2002 Sam Clegg - See COPYING
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Functions for authenticating users of the jukebox.
 * Two authenication methods are available: PAM and custom.
 * Custom is just like the apache .htaccess method (in fact you can use .htaccess 
 * filed with juked).
 *
 * $superduper: auth.c,v 1.9 2002/05/01 11:37:02 samc Exp $
 */
#include "config.h"

#ifdef WITH_PAM
#include <security/pam_appl.h>
#include <security/pam_misc.h>
#endif 

#include "crypt.h"
#include "juked.h"

#ifdef WITH_PAM
static int do_pam_auth   (const char* username, const char* password);
#endif
static int do_custom_auth (const char* username, const char* password);

/** 
 * Perform user authentication.
 * @param username string containing the usename to be authenticated
 * @return 0 on succes, -1 on failure
 */
int
auth_check (const char* username, const char* password)
{
#ifdef WITH_PAM
  if (settings.usepam)
      return do_pam_auth (username, password);
  else
#endif
      return do_custom_auth (username, password);
}

/** 
 * Perform user authentication.
 * @param username string containing the usename to be authenticated
 * @return 0 on succes, -1 on failure
 */
#ifdef WITH_PAM
static int
do_pam_auth (const char* username, const char* password)
{
  struct pam_conv conv = {
      misc_conv,
      NULL
  };

  pam_handle_t* pamh;
  int rtn = pam_start ("juked",username,&conv,&pamh);
  if (rtn != PAM_SUCCESS)
    {
      jdmessage (LOG_ERR, "pam_start failed: %d\n",rtn);
      return -1;
    }
  rtn = pam_authenticate (pamh, 0);  
  if (rtn != PAM_SUCCESS)
    {
      jdmessage (LOG_ERR, "pam_authenticate failed: %d\n",rtn);
      return -1;
    }
  pam_end (pamh, PAM_SUCCESS);
  if (rtn != PAM_SUCCESS)
    {
      jdmessage (LOG_ERR, "pam_end failed: %d\n",rtn);
      return -1;
    }
  return 0;
}
#endif

/** 
 * Perform custom user authentication.
 * This works a lot like apache's .htpasswd system.
 * Only the first 8 chars of the username/password are significant.
 * @param username string containing the usename to be authenticated
 * @return 0 on succes, -1 on failure
 */
#define FIELD_LEN 20
static int
do_custom_auth (const char* username, const char* password)
{
  int i;
  int fatal = 0;
  char c = '\0';
  FILE* pwdfile;
  char tmp[FIELD_LEN];
  char * crypt_pass;

  if (!settings.pwdfile)
    {
      jdmessage (LOG_ERR, "password filename not set\n");
      return -1;
    }
  pwdfile = fopen (settings.pwdfile, "r");
  if (!pwdfile)
    {
      jdmessage (LOG_ERR, "failed to open password file: %s : %s\n", settings.pwdfile, strerror (errno));
      return -1;
    }
  while (1)
    {
      /* read the significant chars of the user name */
      for (i=0; i<FIELD_LEN; i++)
        {
          c = fgetc (pwdfile);
          if (isalnum (c)) tmp[i] = c;
          else if (c == EOF) fatal = 1;
          else break;
        }
      if (fatal) break;

      tmp[i+1] = '\0';
      /* see if we found our user */
      if (!strncmp (username, tmp, i))
        {
          /* skip over the rest of the username till ":" is reached */
          while (c != ':')
            {
              c = fgetc (pwdfile);
              if (c == ':') break;
              else if (c == EOF) fatal = 1;
              else if (!isalnum (c)) fatal = 1;
              if (fatal) break;
            } 
          if (fatal) break;
          
          /* read the significant chars of the password */
          for (i=0; i<FIELD_LEN; i++)
            {
              c = fgetc (pwdfile);
              if (c == EOF || c == '\n') break; 
              tmp[i] = c;
            }
          tmp[i] = '\0';
          crypt_pass = crypt (password, tmp);
          if (!strcmp (crypt_pass, tmp))
            {
              return 0;
            }
          else
            {
              printf ("password incorrect\n");
              return -1;
            }
        }
    }
  /* if we got this far we must have failed to match */
  printf ("username not found\n");
  return -1;
}
