/*
 * juked - a jukebox daemon
 * Copyright (C) 2000-2002 Sam Clegg - See COPYING
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Command functions called from the parser.
 * These almost map 1-to-1 to the jcp commands.
 *
 * $superduper: commands.c,v 1.42 2002/05/02 14:30:49 samc Exp $
 */
#include "config.h"

#include <ctype.h>

#include "juked.h"
#include "commands.h"
#include "playlist.h"
#include "session.h"
#include "database.h"
#include "player.h"

/**
 * Data unique to each parser thread
 */
static pthread_key_t command_key = 0;
xmlDocPtr last_result = 0;

/**
 * Send some XML to the remote socket then free it.  
 * This is used my many of the commend functions
 * @param xml_doc pointer to the XML doc to send & free.
 */
void
xml_output (xmlDocPtr xml_doc)
{
  if (xml_doc)
    {
      int len;
      xmlChar* xml_string;
      if (!command_key) pthread_key_create (&command_key, 0);
      xmlDocDumpFormatMemory (xml_doc, &xml_string, &len, 1);
      jdoutput ("%s.\n", xml_string);
      xmlFree (xml_string);
      last_result = (xmlDocPtr) pthread_getspecific (command_key);
      xmlFreeDoc (last_result);
    }
  else
    {
      jdoutput (".\n");
    }
  pthread_setspecific (command_key, (void*)xml_doc);
}

/**
 * Start playback of a track
 * @param output flag to request output or not
 * @param id the ID of the track to play
 */
void 
cmd_play (int output, int id)
{
  if (db_is_valid_id (id))
    { 
      if (player_start (id))
        {
          if (output) 
            {
              jdoutput ("%serror playing track %d\n", ERROR_STR, id);
              jdoutput (".\n");
            }
        }
      else
        {
          if (output) jdoutput ("%splaying: %d\n", OK_STR, id);
          xml_output (db_lookup (id));
        }
    }
  else
    {
      if (output) 
        {
          jdoutput ("%sbad track id %d\n", ERROR_STR, id);
          jdoutput (".\n");
        }
    }
}

/**
 * Stop all playback
 * @param output flag to request output (freedback).
 */
void 
cmd_stop (int output)
{
  player_stop ();
  jdoutput ("%soutput stopped\n", OK_STR);
  jdoutput (".\n");
}

/**
 * Pause playback
 * @param output flag to request output (freedback).
 */
void
cmd_pause (int output)
{
  int state = player_pause ();
  if (state)
      jdoutput ("%soutput paused\n", OK_STR);
  else
      jdoutput ("%soutput resumed\n", OK_STR);
  jdoutput (".\n");
}

/**
 * Play the next track in the playlist.
 * @param output flag to request output (freedback).
 */
void
cmd_next (int output)
{
  cmd_play (output, plist_next (output));
}

/**
 * Play the previous track in the playlist.
 * @param output flag to request output (freedback).
 */
void
cmd_prev (int output)
{
  cmd_play (output, plist_prev ());
}

/**
 * Display to current playlist.
 */
void
cmd_list ()
{
  jdoutput ("%splaylist follows:\n", OK_STR);
  xml_output (plist_genxml ());
}

/**
 * Clear the playlist
 * @param output flag to request output (freedback).
 */
void
cmd_clearlist (int output)
{
  plist_clear ();
  jdoutput ("%splaylist cleared\n", OK_STR);
  jdoutput (".\n");
}

/**
 * Toggle random play  
 * @param output flag to request output (freedback).
 */
void
cmd_random (int output)
{
  settings.random = !settings.random;
  if (settings.random)
      jdoutput ("%srandom mode : ON\n", OK_STR);
  else
      jdoutput ("%srandom mode : OFF\n", OK_STR);
  jdoutput (".\n");
}

/**
 * Toggle random play  
 * @param output flag to request output (freedback).
 */
void
cmd_autoskip (int output)
{
  settings.autoskip = !settings.autoskip;
  if (settings.autoskip)
      jdoutput ("%sautoskip mode : ON\n", OK_STR);
  else
      jdoutput ("%sautoskip mode : OFF\n", OK_STR);
  jdoutput (".\n");
}

/**
 * Add the given song ID to the playlist.
 * @param output flag to request output (freedback).
 * @param id the ID of the track to be added.
 */
void
cmd_add (int output, int id)
{
  int count = 0;
  xmlDocPtr doc = 0;
  xmlNodePtr node = 0;
  xmlNodePtr child = 0;
  if (id == -1)
    {
      doc = (xmlDocPtr)pthread_getspecific (command_key);
      if (!doc)
        {
          jdoutput ("%sno tracks selected\n", ERROR_STR, id);
        }
      else
        {
          node = xmlDocGetRootElement (doc);
          if (!strcmp (node->name, XML_BASENODE_LIST))
            {
              child = node->children;
              while (child)
                {
                  id = atoi (xmlNodeListGetString (doc, child->xmlChildrenNode, 1));
                  plist_insert (id);
                  count++;
                  child = child->next;
                }
            }
          jdoutput ("%sadded %d tracks\n", OK_STR, count);
        }
    }
  else if (db_is_valid_id (id) == -1)
    {
      jdoutput ("%sbad track id %d\n", ERROR_STR, id);
      jdoutput (".\n");
    }
  else
    {
      jdoutput ("%sadding track %d\n", OK_STR, id);
      plist_insert (id);
    }
  jdoutput (".\n");
}

/**
 * Show track information
 * @param id show info on track with this ID.
 */
void
cmd_info (int id)
{
  xmlDocPtr doc = db_lookup (id);
  if (doc)
    {
      jdoutput ("%sresults follow:\n", OK_STR);
      xml_output (doc);
    }
  else
    {
      jdoutput ("%sno track info found (%d)\n", ERROR_STR, id);
      jdoutput (".\n");
    }
}

/**
 * Remove a track form the current playlist
 * @param output flag to request output (freedback).
 * @param id thrack ID to remove.
 */
void
cmd_remove (int output, int id)
{
  if (db_is_valid_id (id) == -1)
    {
      jdoutput ("%sbad track id %d\n", ERROR_STR, id);
    }
  else
    {
      jdoutput ("%sremoving track %d\n", OK_STR, id);
      plist_remove (id);
    }
  jdoutput (".\n");
}

/**
 * End this session.
 * The parser actual terminates the session.
 */
void
cmd_quit (int output)
{
  jdoutput ("%sGoodbye\n", OK_STR);
  jdoutput (".\n");
}

/**
 * Show the current status. i.e. current track info
 */
void
cmd_status ()
{
  int id;
  id = player_get_current_id ();
  jdoutput ("%sserver status follow:\n", OK_STR);
  jdoutput ("juked version    : %s\n", VERSION);
  jdoutput ("servername       : %s\n", localhost);
  jdoutput ("uptime (secs)    : %d\n", time (0) - startup_time);
  jdoutput ("connected clients: %d\n", get_num_sessions());
  jdoutput ("tracks played    : %d\n", player_get_playcount ());
  jdoutput ("now playing      : %d\n", id);
  jdoutput (".\n");
}

/**
 * Show the history
 */
void
cmd_history ()
{
  jdoutput ("%shistory follows:\n", OK_STR);
  player_show_history ();
  jdoutput (".\n");
}

/**
 * Show the current juked version.
 */
void
cmd_version ()
{
  jdoutput ("%sJuked v%s.\n", OK_STR, VERSION);
  jdoutput (".\n");
}

/**
 * Store the username and await password command
 */
void
cmd_user (char* username)
{
  struct session_data_t* session_data = pthread_getspecific (session_key);
  strncpy (session_data->username, username, _POSIX_LOGIN_NAME_MAX);
  jdoutput ("%sUsername set. Awaiting pass.\n", OK_STR);
  jdoutput (".\n");
}

/**
 *
 */
void
cmd_pass (char* password)
{
  struct session_data_t* session_data = pthread_getspecific (session_key);
  jdoutput ("%sGot pass %s for user %s\n", OK_STR, password, session_data->username);
  jdoutput (".\n");
}

/**
 * Show the help message
 * This lists all the commands available to the user.
 */
void
cmd_help ()
{
  jdoutput ("%sUsage message follows:\n", OK_STR);
  jdoutput ("This is juked v%s.\n", VERSION);
  jdoutput ("Available commands:\n");
  jdoutput ("  play <trackid>       - start a track playing.\n");
  jdoutput ("  stop                 - stop output.\n");
  jdoutput ("  pause                - pause output.\n");
  jdoutput ("  ladd <trackid>       - add a track to the current playlist.\n");
  jdoutput ("  lrm <trackid>        - remove a track to the current playlist.\n");
  jdoutput ("  llist                - display the contents of the current playlist.\n");
  jdoutput ("  random               - toggle random mode.\n");
  jdoutput ("  autoskip             - toggle autoskip mode.\n");
  jdoutput ("  find [<field>] <txt> - find tracks containing given text.\n");
  jdoutput ("  info <trackid>       - display info on a track.\n");
  jdoutput ("  exit                 - exit juked.\n");
  jdoutput ("  help                 - display this help message.\n");
  jdoutput (".\n");
}

/**
 * Search the database.
 * @param field the database field to search in (may by null)
 * @param string the string to search for.
 */
void
cmd_find (char* field, char* string)
{
  assert (string);
  jdoutput ("%ssearch results follow:\n", OK_STR);
  xml_output (db_search (field, string));
}

/**
 * List all the artists in the database
 */
void
cmd_artists ()
{
  jdoutput ("%sthe following artists exist in the database:\n", OK_STR);
  xml_output (db_list (0));
}
