/*
 * juked - a jukebox daemon
 * Copyright (C) 2000 Sam Clegg - See COPYING
 *  
 * Internal database for functions for use with juked.
 *
 * $superduper: db_builtin.c,v 1.16 2002/05/02 10:38:56 samc Exp $
 */
#include <dirent.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <regex.h>
#include <id3tag.h>

#include "../juked.h"
#include "../database.h"

#define DATA_FILENAME "juked.data.xml.gz"

#define DIR_SEPARATOR '/'

/**
 * The buildin database is a linked list of these.
 */
typedef struct entry* entryptr;

struct entry
{
  int id;
  char filename[_POSIX_PATH_MAX];
  char *artist;
  char *title;
  char *album;
  char *comment;
  int tracknum;
  entryptr next;
};

/**
 * The singleton database.
 */
struct
{
  entryptr first;   /* the first track      */
  entryptr last;    /* the last track       */
  entryptr *index;  /* track index          */
  int count;        /* number of tracks     */
  int errors;       /* number of bad tracks */
} data;

/**
 * Internal function for reading ID3 tags.
 */
static int
entry_init (const char *filename, entryptr node)
{
  struct id3_tag *tag;
  struct id3_file *id3file;
  struct id3_frame const *frame;
  union id3_field const *field;
  memset (node, 0, sizeof (struct entry));
  node->tracknum = -1;
  id3file = id3_file_open (filename, ID3_FILE_MODE_READONLY);
  tag = id3_file_tag (id3file);
  frame = id3_tag_findframe (tag, ID3_FRAME_ARTIST, 0);
  if (frame)
    {
      field = &frame->fields[1];
      node->artist = id3_ucs4_latin1duplicate (id3_field_getstrings (field, 0));
    }
  frame = id3_tag_findframe (tag, ID3_FRAME_TITLE, 0);
  if (frame)
    {
      field = &frame->fields[1];
      node->title = id3_ucs4_latin1duplicate (id3_field_getstrings (field, 0));
    }
  frame = id3_tag_findframe (tag, ID3_FRAME_TRACK, 0);
  if (frame)
    {
      /*char * str;
      field = &frame->fields[1];
      str = id3_ucs4_latin1duplicate (id3_field_getstrings (field, 0));
      printf ("%s\n", str);
      node->tracknum = atoi (str);
      free (str);
      */
    }
  frame = id3_tag_findframe (tag, ID3_FRAME_ALBUM, 0);
  if (frame)
    {
      field = &frame->fields[1];
      node->album = id3_ucs4_latin1duplicate (id3_field_getstrings (field, 0));
    }
  id3_file_close (id3file);
  return 0;
}

/**
 * Internal function for adding nodes to the database.
 */
static int
add_entry (entryptr new_node)
{
  new_node->id = ++data.count;
  new_node->next = 0;
  if (!data.first)
    {
      data.first = new_node;
    }
  else
    {
      data.last->next = new_node;
    }
  data.last = new_node;
  return 0;
}

/**
 * Internal utility function to generate a XML data representing a single
 * data node.
 */
static xmlNodePtr
gen_xml (entryptr data_node)
{
  char numstr[10];    
  xmlNodePtr root = 0;
  xmlNodePtr node = 0;

  root = xmlNewNode (0, XML_BASENODE_INFO);

  node = xmlNewChild (root, 0, XML_TAG_ARTIST, 0);
  xmlAddChild (node, xmlNewText (data_node->artist));

  node = xmlNewChild (root, 0, XML_TAG_ALBUM, 0);
  xmlAddChild (node, xmlNewText (data_node->album));

  node = xmlNewChild (root, 0, XML_TAG_TITLE, 0);
  xmlAddChild (node, xmlNewText (data_node->title));

  node = xmlNewChild (root, 0, XML_TAG_TRACKNUM, 0);
  memset (numstr, 0, 10);
  snprintf (numstr, 9, "%d", data_node->tracknum);
  xmlAddChild (node, xmlNewText (numstr));

  node = xmlNewChild (root, 0, XML_TAG_TRACKID, 0);
  memset (numstr, 0, 10);
  snprintf (numstr, 9, "%d", data_node->id);
  xmlAddChild (node, xmlNewText (numstr));

  node = xmlNewChild (root, 0, "filename", 0);
  xmlAddChild (node, xmlNewText (data_node->filename));

  if (data_node->comment)
    {
      node = xmlNewChild (root, 0, XML_TAG_COMMENT, 0);
      xmlAddChild (node, xmlNewText (data_node->comment));
    }
  return root;
}

/**
 * Internal recursive function to find all the mp3s in a given 
 * direcotry and add them to the database.
 */
static int 
create_database (const char *directory)
{
  int rtn = 0;
  entryptr new_node;
  struct dirent *entry;
  struct stat stat_info;
  DIR *dir;
  char cwd[_POSIX_PATH_MAX+1];
  char previous_cwd[_POSIX_PATH_MAX+1];

  memset (cwd, 0, _POSIX_PATH_MAX+1);
  memset (previous_cwd, 0, _POSIX_PATH_MAX+1);
  dir = opendir (directory);
  if (!dir) 
    {
      jdmessage (LOG_ERR, "db_buildin : %s", directory, strerror (errno));
      return -1;
    }
  if (!getcwd (previous_cwd, _POSIX_PATH_MAX))
    {
      jdmessage (LOG_ERR, "db_buildin : getcwd: %s", strerror (errno));
      return -1;
    }
  if (chdir (directory) == -1)
    {
      jdmessage (LOG_ERR, "db_buildin : %s: %s", 
                           directory, strerror (errno));
      return -1;
    }
  if (!getcwd (cwd, _POSIX_PATH_MAX))
    {
      jdmessage (LOG_ERR, "db_buildin : getcwd : %s", strerror (errno));
      return -1;
    }

  while ((entry = readdir (dir)) && rtn == 0)
    {
      if (stat (entry->d_name, &stat_info) == -1)
        {
          jdmessage (LOG_ERR, "db_buildin : %s : %s", entry->d_name, strerror (errno));
        }
      else if (S_ISDIR(stat_info.st_mode))
        {
          if (entry->d_name[0] != '.')
            {
              /* recurse into driectory */
              rtn = create_database (entry->d_name);
            }
        }
      else
        {
          char* extension = entry->d_name + strlen (entry->d_name) - 4;
          if (!strncasecmp (extension, ".mp3", 4))
            {
              new_node = (entryptr) malloc (sizeof (struct entry));
              if (!new_node) rtn = -1;
              if (entry_init (entry->d_name, new_node) == -1)
                {
                  jdmessage (LOG_DEBUG, "db_buildin : failed to read ID3 tag: %s", entry->d_name);
                  free (new_node);
                  continue;
                }
              //assert (!strncmp (settings.mp3root,cwd,strlen (settings.mp3root)));
              strncpy (new_node->filename, cwd+strlen (settings.mp3root), _POSIX_PATH_MAX);
              strncat (new_node->filename, "/", 1);
              strncat (new_node->filename, entry->d_name, _POSIX_PATH_MAX);
              add_entry (new_node);
            }
        }
    }
  closedir (dir);
  if (chdir (previous_cwd) == -1)
    {
      jdmessage (LOG_ERR, "db_buildin : %s : %s", previous_cwd, strerror(errno));
      return -1;
    }
  return rtn;
}

/**
 * Load XML database from file
 */
static int
load_database (const char* filename)
{
  char* string;
  entryptr new_data;
  xmlNodePtr xml_entry;
  xmlNodePtr xml_node;
  xmlDocPtr doc = xmlRecoverFile (filename); 
  if (!doc) return -1;
  xml_entry = xmlDocGetRootElement (doc);
  /* Check we have a data file */
  if (strcmp (xml_entry->name, XML_BASENODE_DATA))
    {
      xmlFreeDoc (doc);
      return -1;
    }
  xml_entry = xml_entry->children;
  while (xml_entry != NULL)
    {
      if (strcmp (xml_entry->name, XML_BASENODE_INFO))
        {
          xmlFreeDoc (doc);
          return -1;
        }
      new_data = (entryptr) malloc (sizeof (struct entry));
      if (!new_data) return -1;
      memset (new_data, 0, sizeof (struct entry));
      add_entry (new_data);
      xml_node = xml_entry->children;
      while (xml_node != NULL)
        {
          if (!xmlIsBlankNode (xml_node))
            {
              string = xmlNodeListGetString (doc, xml_node->xmlChildrenNode, 1);
              if (string) 
                {
                  if (!strcmp (xml_node->name, "id"))
                    {
                      new_data->id = atoi (string);
                    }
                  else if (!strcmp (xml_node->name, "tracknum"))
                    {
                      new_data->tracknum = atoi (string);
                    }
                  else if (!strcmp (xml_node->name, "artist"))
                    {
                      strcpy (new_data->artist, string);
                    }
                  else if (!strcmp (xml_node->name, "album"))
                    {
                      strcpy (new_data->album, string);
                    }
                  else if (!strcmp (xml_node->name, "title"))
                    {
                      strcpy (new_data->title, string);
                    }
                  else if (!strcmp (xml_node->name, "comment"))
                    {
                      printf("found comment\n");
                      printf("%s\n", string);
                      strcpy (new_data->comment, string);
                    }
                  else if (!strcmp (xml_node->name, "filename"))
                    {
                      strcpy (new_data->filename, string);
                    }
                  free (string);
                }
            }
          xml_node = xml_node->next;
        }
      xml_entry = xml_entry->next;
    }
  xmlFreeDoc (doc);
  return 0;
}

/**
 * Dump the database to a file
 */
static int
save_database (const char* filename)
{
  int rtn = 0;
  xmlDocPtr doc;
  entryptr  node;
  doc       = xmlNewDoc ("1.0");
  doc->charset = XML_CHAR_ENCODING_8859_1;
  xmlSetDocCompressMode (doc, 9);
  xmlSetCompressMode (9);
  xmlIndentTreeOutput = 1;
  doc->children = xmlNewDocNode (doc, 0, XML_BASENODE_DATA, 0);
  node = data.first;
  for (node = data.first; node; node = node->next)
    {
      xmlAddChild (xmlDocGetRootElement (doc), gen_xml (node));
    }
  if (xmlSaveFormatFile (filename, doc, 1) == -1)
    {
      jdmessage (LOG_ERR, "db_buildin : failed to write XML data : %s : %s", filename, strerror (errno));
      rtn = -1;
    } 
  xmlFreeDoc (doc);
  return rtn;
}

/**
 * @return 0 on success, -1 on failure
 */
int 
init_database_builtin ()
{
  entryptr node;
  struct stat statinfo;
  char datafile[_POSIX_PATH_MAX];
  int count  = 0;
  int len    = 0;
  int loaded = 0;
  data.first = 0;
  data.last  = 0;
  data.count = 0;
  data.index = 0;
  data.errors= 0;

  jdmessage (LOG_DEBUG, "init builtin database");
  memset (datafile, 0, _POSIX_PATH_MAX);
  strncpy (datafile, settings.mp3root, _POSIX_PATH_MAX-1);
  len = strlen(datafile);
  if (datafile[len-1] != DIR_SEPARATOR)
    {
      if (len > _POSIX_PATH_MAX - 3)
        {
          jdmessage (LOG_ERR, "overflowed _POSIX_PATH_MAX");
          return -1;
        }
      datafile[len] = DIR_SEPARATOR;
    }
  strncat (datafile, DATA_FILENAME, _POSIX_PATH_MAX - strlen (datafile) - 1);
  if (stat (datafile, &statinfo) != -1)
    {
      jdmessage (LOG_DEBUG, "db_builtin : loading XML : %s", datafile);
      if (load_database (datafile) == -1)
        {
          jdmessage (LOG_ERR, "db_builtin : failed to read XML data : %s", datafile);
        }
      else
        {
          loaded = 1;
          jdmessage (LOG_DEBUG, "db_builtin : read %d records", data.count);
        }
    }
  if (!loaded)
    {
      if (create_database (settings.mp3root) == -1) 
        {
          jdmessage (LOG_ERR, "db_buildin : failed to build track database : %s", settings.mp3root);
          return -1;
        }
      jdmessage (LOG_DEBUG, "db_buildin : found %d tracks", data.count);
      if (data.errors)
        {
          jdmessage (LOG_DEBUG, "db_buildin : omitted %d tracks with bad ID tags", data.errors);
        }
      jdmessage (LOG_DEBUG, "db_builtin : writing : %s", datafile);
      count = 0;
      save_database (datafile);
    }
  jdmessage (LOG_INFO, "db_builtin : building track index");
  data.index = (entryptr*) malloc ((data.count + 1) * sizeof (entryptr));
  memset (data.index, 0, (data.count + 1) * sizeof (entryptr));
  node = data.first;
  while (node)
    {
      if (data.index[node->id] != 0)
        {
          jdmessage (LOG_ERR, "db_builtin : error building index : %d : %d", node->id, data.index[node->id]);
          return -1;
        }
      data.index[node->id] = node;
      node = node->next;
      count++;
    }
  return 0;
}

/**
 * @return 0 on success, -1 on failure
 */
int
term_database_builtin ()
{
  entryptr node = data.first;
  entryptr next;
  jdmessage (LOG_DEBUG, "term_database_builtin");
  while (node)
    {
      next = node->next;
      free (node);
      node = next;
    }
  free (data.index);
  return 0;
}

/**
 * Internal utility function..
 */
static entryptr
get_entry (int track_id)
{
  if (track_id > data.count || track_id < 1) 
    {
      return 0;
    }
  else 
    {
      return data.index[track_id];
    }
}

/**
 * Get the filename associated with a given ID
 * If the filename is longer than the given length then we truncate it
 * @return 0 on success, -1 on failure
 */
int 
db_get_filename_builtin (int track_id, char* filename)
{
  entryptr node = get_entry (track_id);
  if (node)
    {
      strncpy (filename, node->filename, strlen (node->filename) + 1);
      return 0;
    }
  else return -1;
}

/**
 * Return non-zero if there exists a song in the
 * database with the given ID
 * @return 1 on success, 0 on failure
 */
int
db_is_valid_id_builtin (int track_id)
{
  entryptr node = get_entry (track_id);
  if (node) return 1;
  else      return 0;
}

/**
 * Return a random int between 0 and the number of tracks in the database.
 * @return valid track_id on success, -1 on failure
 */
int
db_get_random_id_builtin ()
{
  return data.count * ((float)rand () / (float)RAND_MAX);
}

/**
 * Search the database for records containing a given string is a given field.
 * If no field is given then assume *
 * Build XML tree of resulting IDs
 * @param field the field the seach within. May be NULL
 * @param string the string to search for. Cannot be NULL
 * @return a pointer to a valid XML document, NULL on failure
 */
xmlDocPtr
db_search_builtin (char *field, char *string)
{
  entryptr db_node;      /* node for intertion through database */
  regex_t regexp;        /* the compiled regexp */
  regmatch_t matches[1]; /* we are only ever looking for one match */
  char buff[256];        /* error string */
  xmlDocPtr  doc  = 0;   /* reults go in am XML doc */
  xmlNodePtr node = 0;
  int rtn = regcomp (&regexp, string, REG_ICASE);
  if (rtn != 0)
    {
      jdmessage (LOG_ERR, "unable to compile regex: %d: %s", rtn, string);
      return 0;
    }
  doc       = xmlNewDoc ("1.0");
  doc->children = xmlNewDocNode (doc, 0, XML_BASENODE_LIST, 0);

  db_node = data.first;
  while (db_node)
    {
      strncpy (buff, db_node->artist, 256);
      rtn = regexec (&regexp, buff, 1, matches, 0);
      switch (rtn)
        {
        case 0:
          node = xmlNewChild (xmlDocGetRootElement (doc), 0, XML_TAG_TRACKID, 0);
          snprintf (buff, 256, "%d", db_node->id);
          xmlAddChild (node, xmlNewText (buff));
          break;
        case REG_NOMATCH:
          break;
        default:
          regerror (rtn, &regexp, buff, 256);
          jdmessage (LOG_ERR, "regexec error: %s: %s", buff, string);
          break;
        }
      db_node = db_node->next;
    }
  regfree (&regexp);
  if (node) return doc;
  xmlFreeDoc (doc);
  return 0;
}

/**
 * Look up an ID in the database and build an XML tree of results
 * @param id the ID to lookup
 * @return a pointer to a valid XML document NULL on failure
 */
xmlDocPtr
db_lookup_builtin (int track_id)
{
  xmlDocPtr  doc  = 0;
  entryptr   node = 0;
  doc    = xmlNewDoc ("1.0");
  node   = get_entry (track_id);
  if (!node)
    {
      xmlFreeDoc (doc);
      return 0;
    }
  xmlNewNode (0, "hello");
  xmlDocSetRootElement (doc, gen_xml (node));
  return doc;
}

/**
 * 
 */
xmlDocPtr 
db_list_builtin (int trackid)
{
  //entryptr node = data.first;
  return 0;
}
