/*
 * juked - a jukebox daemon
 * Copyright (C) 2000 Sam Clegg - See COPYING
 *  
 * Funcions for accessing the database component of the jukebox
 * 
 * $superduper: db_mysql.c,v 1.4 2002/05/02 09:57:53 samc Exp $
 */
#include "../juked.h"
#include "../database.h"

#ifdef HAVE_MYSQL
#include <mysql/mysql.h>

/**
 * The mysql connection info.
 * This is global within this file and persistent between
 * function calls.
 */
static MYSQL* mysql      = 0;
static MYSQL* connection = 0;

/**
 * opens the connection to the database
 * @return 0 on success, -1 on failure
 */
int 
init_database_mysql()
{
  assert(!connection);
  mysql = (MYSQL*)malloc(sizeof(MYSQL));
  mysql_init(mysql);
  jdmessage(LOG_INFO,"opening database \"%s\" on \"%s\"\n", settings.db_name, settings.db_server); 
  connection = mysql_real_connect(mysql,
      settings.db_server,
      settings.db_user,
      settings.db_pass,
      settings.db_name,0,0,0);
  if (!connection)
    {
      jdmessage(LOG_ERR,"failed to open database : %s\n", mysql_error(mysql)); 
      return -1;
    }
  return 0;
}

/**
 * Wrapper around mysql_query
 * @return 0 on success, -1 on failure
 */
static int 
safe_query(char* query)
{
  if (!connection) return -1;
  if (mysql_query(connection,query))
    {
      jdmessage(LOG_ERR,"failed query : %s\n",mysql_error(connection));
      return -1;
    }
  return 0;
}

/**
 * Close to connextion to the database
 * @return 0 on success, -1 on failure
 */
int
term_database_mysql()
{
  if (connection)
    {
      mysql_close(connection);
      return 0;
    }
  else
      return -1;
}

/**
 * Get the filename associated with a given ID
 * @param filename file name to fill. Should be big enough to hold at most _POSIX_PATH_MAX
 * @return 0 on success, -1 on failure
 */
int 
db_get_filename_mysql(int track_id, char* filename)
{
  MYSQL_ROW row;
  MYSQL_RES *res  = 0;
  int    len      = 0;
  int    num_rows = 0;
  char * pos      = 0;
  char   query[]  = "SELECT filename FROM tracks WHERE ID=     ";

  len = strlen(query);
  pos = &query[len-5];
  snprintf(pos,5,"%d",track_id);
  if (safe_query(query)) return -1;
  res = mysql_store_result(connection);
  num_rows = mysql_num_rows(res);
  if (num_rows == 0) return -1;
  if (num_rows > 1)  jdmessage(LOG_DEBUG,"mulpiple matches?!?! (%d)\n",track_id);

  row = mysql_fetch_row(res);
  strncpy(filename,row[0],_POSIX_PATH_MAX);
  mysql_free_result(res);
  return 0;
}

/**
 * Return non-zero if there exists a song in the
 * database with the given ID
 * @return 1 on success, 0 on failure
 */
int
db_is_valid_id_mysql(int track_id)
{
  MYSQL_RES *res  = 0;
  int    len      = 0;
  int    num_rows = 0;
  char * pos      = 0;
  char   query[]  = "SELECT filename FROM tracks WHERE ID=     ";

  len = strlen(query);
  pos = &query[len-5];
  snprintf(pos,5,"%d",track_id);
  if (safe_query(query)) return 0;
  res = mysql_store_result(connection);
  num_rows = mysql_num_rows(res);
  if (num_rows == 0) return 0;
  if (num_rows > 1) jdmessage(LOG_DEBUG,"mulpiple matches?!?! (%d)\n",track_id);

  mysql_free_result(res);
  return 1;
}

/**
 * Return a random int between 0 and the number of tracks in the database.
 * [TODO] After choosing random number we should check
 * if it is valid (e.g. it may have been deleted from the db)
 * @return valid track_id on success, -1 on failure
 */
int
db_get_random_id_mysql()
{
  int    rtn      = -1;
  MYSQL_RES *res  = 0;
  MYSQL_ROW row;
  char query[] = "SELECT COUNT(*) FROM tracks";

  if (safe_query(query)) return -1;
  res = mysql_store_result(connection);
  row = mysql_fetch_row(res);
  rtn = atoi(row[0]);
  mysql_free_result(res);
  rtn *= ((float)rand() / (float)RAND_MAX);
  return rtn;
}

static const char* DEFAULT_FIELD = "artist";
/**
 * Search the database for records containing a given string is a given field.
 * If no field is given then assume *
 * Build XML tree of resulting IDs
 * @param field the field the seach within. May be NULL
 * @param string the string to search for. Cannot be NULL
 * @return a pointer to a valid XML document, NULL on failure
 */
xmlDocPtr
db_search_mysql(char* field, char* string)
{
  xmlDocPtr  doc  = 0;
  xmlNodePtr node = 0;
  xmlNsPtr   ns   = 0;
  MYSQL_RES *res  = 0;
  MYSQL_ROW row;

  const char * field_string = field ? field : DEFAULT_FIELD;
  char query[256];
  char base_query[] = "SELECT ID FROM tracks WHERE";
  snprintf(query,256,"%s %s LIKE \"%%%s%%\"", base_query, field_string, string);

  jdmessage(LOG_DEBUG, "looking for: %s in: %s\n", string, field_string); 
  
  if (safe_query(query)) return 0;
  res = mysql_store_result(connection);

  doc       = xmlNewDoc     ("1.0");
  ns        = xmlNewGlobalNs(doc, XML_URI, XML_NAMESPACE);
  doc->root = xmlNewDocNode (doc, ns, XML_BASENODE_LIST, 0);
  while ((row = mysql_fetch_row(res)))
    {
      node = xmlNewChild(doc->root, ns, XML_TAG_TRACKID, 0);
      xmlAddChild(node, xmlNewText(row[0]));
    }

  /*
     jdoutput("%sResults follow. Found %d matches\n",OK_STR,num_rows);
     jdoutput(" ID     Artist           Album            Track Name\n");
     while ((row = mysql_fetch_row(res)))
     {
     jdoutput("[%5.5s] %-20.20s - %-18.18s - %-25.25s\n",
     row[0], row[1], row[2], row[3]);
     }
     jdoutput(".\n");
   */
  mysql_free_result(res);
  xmlFreeNs(ns);
  return doc;
}

/**
 * Look up an ID in the database and build an XML tree of results
 * @param id the ID to lookup
 * @return a pointer to a valid XML document NULL on failure
 */
xmlDocPtr
db_lookup_mysql(int track_id)
{
  xmlDocPtr  doc      = 0;
  xmlNodePtr node     = 0;
  xmlNsPtr   ns       = 0;
  MYSQL_RES *res      = 0;
  MYSQL_FIELD *fields = 0;
  MYSQL_ROW row;
  int    i;
  int    num_fields   = 0;
  int    len          = 0;
  int    num_rows     = 0;
  char * pos          = 0;
  char   query[]      = "SELECT * FROM tracks WHERE ID=     ";

  len = strlen(query);
  pos = &query[len-5];
  snprintf(pos,5,"%d",track_id);
  if (safe_query(query)) return 0;
  res = mysql_store_result(connection);
  num_rows = mysql_num_rows(res);
  if (num_rows == 0) return 0;
  if (num_rows > 1)  jdmessage(LOG_DEBUG,"mulpiple matches!?! (%d)\n",track_id);
  row = mysql_fetch_row(res);
  doc       = xmlNewDoc     ("1.0");
  ns        = xmlNewGlobalNs(doc, XML_URI, XML_NAMESPACE);
  doc->root = xmlNewDocNode (doc, ns, XML_BASENODE_INFO, 0);
  num_fields = mysql_num_fields(res);
  fields = mysql_fetch_fields(res);
  for (i=0; i<num_fields; i++)
    {
      node = xmlNewChild(doc->root, ns, fields[i].name, 0);
      xmlAddChild(node, xmlNewText(row[i]));
    }
  xmlFreeNs(ns);
  mysql_free_result(res);
  return doc;
}

#endif /* HAVE_MYSQL */
