/*
 * juked - a jukebox daemon 
 * Copyright (C) 2000 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *    
 * libmad playback function headers.
 *
 * $superduper: libmad.c,v 1.10 2002/05/02 10:40:33 samc Exp $
 */
#include "config.h"

#ifdef HAVE_LIBMAD

#include "juked.h"
#include <mad.h>
#include <id3tag.h>

struct 
{
  struct mad_stream stream;
  struct mad_frame  frame;
  struct mad_synth  synth;
  char  *buffer;           /* buffer of encoded mp3 data */
  int    size;             /* the size of buffer */
  int    pos;              /* current position in buffer */    
  int    framecount;       /* number of frames decoded */
  int    leftover_pcm;     /* flag */
} mad;

int
mad_init ()
{
  /* init mad stuff */
  mad.leftover_pcm = 0;
  mad.framecount   = 0;
  mad.buffer       = 0;
  mad.size         = 0;
  mad.pos          = 0;
  mad_synth_init (&mad.synth);
  mad_frame_init (&mad.frame);
  mad_stream_init (&mad.stream);
  return 0;
}

int
mad_term()
{
  mad_synth_finish (&mad.synth);
  mad_frame_finish (&mad.frame);
  mad_stream_finish (&mad.stream);
  return 0;
}

/**
 * Scale PCM data
 */
static inline
signed int scale (mad_fixed_t sample)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - 16));

  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;

  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - 16);
}

/**
 * Write PCM data to an output string
 */
static int
write_pcm (struct mad_pcm *pcm, struct mad_header* header, char* outbuf, int len)
{
  int i = 0;
  unsigned int nsamples;
  mad_fixed_t const *left_ch, *right_ch;

  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];

  /* Return now is there is not enough space for the entire PCM sample */
  if (MAD_NCHANNELS(header) == 2)
    {
      if ((nsamples * 4) > len) return -1;
    }
  else
    {
      if ((nsamples * 2) > len) return -1;
    }

  while (nsamples--) 
    {
      signed int sample;
      /* output sample(s) in 16-bit signed little-endian PCM */
      sample = scale (*left_ch++);
      outbuf[i++] = (sample >> 0) & 0xff;
      outbuf[i++] = (sample >> 8) & 0xff;

      if (MAD_NCHANNELS(header) == 2)
        {
          sample = scale (*right_ch++);
          outbuf[i++] = (sample >> 0) & 0xff;
          outbuf[i++] = (sample >> 8) & 0xff;
        }
    }
  return i;
}

/**
 * Add some data to the mad buffer
 */
static inline int
data_add (char *bufin, int inlength)
{
  int   new_size   = 0;
  int   remainder  = mad.size - mad.pos;
  char* old_buffer = mad.buffer;

  assert (remainder >= 0);

  /* create new buffer */
  new_size = inlength + remainder;
  mad.buffer = (char *) malloc (new_size * sizeof (char));

  /* copy in old data if there was any */
  memcpy (mad.buffer, old_buffer + mad.pos, remainder);

  /* free OLD buffer */
  free (old_buffer);

  /* copy in new data */
  memcpy (mad.buffer + remainder, bufin, inlength);
  mad.pos = 0;
  mad.size = new_size;
  return 0;
}

int
mad_decode (char *bufin, char *bufout, int inlen, int outlen, int *result)
{
  int rtn     = 0;
  int written = 0;
  int tmp     = 0;

  /* if we have any new data then add it to the stream */
  if (bufin)
    {
      data_add (bufin, inlen);
    }

  if (mad.leftover_pcm)
    {
      written = write_pcm (&mad.synth.pcm, &mad.frame.header, bufout, outlen);
      mad.leftover_pcm = 0;
    }

  /* point the mad stream at the local in buffer */
  mad_stream_buffer (&mad.stream, mad.buffer + mad.pos, mad.size - mad.pos);

  /* decode and write the outbuf until it is full */
  while (1)
    {
      if (mad_header_decode(&mad.frame.header, &mad.stream) == -1)
        {
          if (!MAD_RECOVERABLE(mad.stream.error)) 
            {
              rtn = -1;
              break;
            }
          if (mad.stream.error == MAD_ERROR_LOSTSYNC)
            {
              /* ignore LOSTSYNC due to ID3 tags */
              int tagsize = id3_tag_query (mad.stream.this_frame, &mad.stream.bufend - &mad.stream.this_frame);
              if (tagsize > 0)
                { 
                  mad_stream_skip (&mad.stream, tagsize);
                  continue;
                }
            }
          jdmessage (LOG_ERR, "libmad : error decoding header at frame %d: %s", 
              mad.framecount, 
              mad_stream_errorstr (&mad.stream));
          continue;
        }
      if (mad_frame_decode(&mad.frame, &mad.stream) == -1)
        {
          if (!MAD_RECOVERABLE(mad.stream.error))
            {
              jdmessage (LOG_ERR, "libmad : error decoding frame at frame %d: %s", 
                  mad.framecount,
                  mad_stream_errorstr (&mad.stream));
              break;
            }
          jdmessage (LOG_ERR, "libmad : error decoding frame %d: %s", 
                     mad.framecount, 
                     mad_stream_errorstr (&mad.stream));
          continue;
        }
      mad.framecount++;
      mad_synth_frame (&mad.synth, &mad.frame);
      mad_stream_sync (&mad.stream);
      /* increment the position in the input buffer */
      tmp = write_pcm (&mad.synth.pcm, &mad.frame.header, bufout+written, outlen-written);
      if (tmp == -1) 
        {
          mad.leftover_pcm = 1;
          break;
        }
      written += tmp;
    }
  *result = written;
  mad.pos += mad.stream.next_frame - mad.stream.buffer;
  return rtn;
}

#endif /* HAVE_LIBMAD */
