/*
 * juked - a jukebox daemon
 * Copyright (C) 2000-2002 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  
 * The main juked control functions
 * This file contains all startup and shutdown functions as well
 * as funstion for thread control
 *
 * $superduper: juked.c,v 1.33 2002/05/02 10:58:33 samc Exp $
 */
#include "config.h"

#include <stdio.h>

#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif /* HAVE_SIGNAL_H */

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */

#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif /* HAVE_SYS_IOCTL_H */

#include "juked.h"
#include "socket_io.h"
#include "commands.h"
#include "playlist.h"
#include "player.h"
#include "session.h"
#include "database.h"

/* the list of active sessions */
static struct session_data_t *sessions = 0;

/* data unique to each session thread */
pthread_key_t session_key;

/* the mutex for accessing the list */
static pthread_mutex_t list_mutex;

/* signal when connection is estabilished */
static pthread_cond_t  connection_signal;
static pthread_mutex_t connection_mutex;

/**
 * Forward declaration of some impending functions
 */
static int   init_signals        (void);
static void  jdshutdown          (int rtncode);
static void *main_session_thread (void *session);
static void  handle_connections  (void);

time_t startup_time;

/**
 * A couple of functions not listed in the headers 
 */
extern int yyparse (void);
extern int parse_opts (int argc, char** argv);


/**
 * Initialise all the subsystems.  This is called on startup
 * and on SIGHUP
 * @see init_signals
 * @see init_plist
 * @see init_settings
 * @see init_mp3
 * @see init_database
 */
void initialise ()
{

  jdmessage (LOG_INFO, "%s v%s - initilising", PACKAGE_NAME, PACKAGE_VERSION); 
  if (init_plist ()     == -1) jdshutdown (-4);
  if (init_player ()    == -1) jdshutdown (-6);
  if (init_database ()  == -1) jdshutdown (-7); 
  if (init_socket_io () == -1) jdshutdown (-8);
  jdmessage (LOG_INFO, "initialization complete"); 
}

/**
 * The main function (entry point)
 * @see parse_opts
 */
int 
main (int argc, char** argv)
{
  int fd;
  startup_time = time (0);
  openlog (PACKAGE_NAME, LOG_PID, LOG_DAEMON);
  if (init_signals ()   == -1) jdshutdown (-3);
  if (init_settings ()  == -1) jdshutdown (-5);
  if (parse_opts (argc, argv)) jdshutdown (-2);

  /* become a daemon */
  if (settings.daemon)
    {
      daemon (0, 0);
      if ((fd = creat(PID_FILE, O_WRONLY)) >= 0)
        {
          char pid[10];
          snprintf (pid, 10, "%d\n", getpid ());
          write (fd, pid, strlen(pid));
          close(fd);  
        }
      else
        {
          jdmessage (LOG_ERR, "%s : %s\n", PID_FILE, strerror(errno));
        }
    }

  initialise ();
  handle_connections ();
  jdshutdown (0);
  return 0;
}

/**
 * Output a debug string at a given debug level
 * This function outputs a string to stdout or syslog depending on 
 * the current level of debugging.  
 * @param fmt the message format (as in printf)
 * @param ... the positions paramter for the message
 */
void        
jdmessage (int log_level, char* fmt, ...)
{
  char string[256];
  va_list args;
  va_start (args, fmt);

  if (settings.log_level >= log_level) 
     {
       if (settings.use_syslog)
         {
           vsnprintf (string, 256, fmt, args);
           syslog (log_level, string);
         }
       fprintf (stderr,"%s : ", PACKAGE_NAME);
       vfprintf (stderr, fmt, args);
       fprintf (stderr,"\n");
     }
  va_end (args);
}

/**
 * @param fmt the message format (as in printf)
 * @param ... the positions paramter for the message
 * @see socket_io_write
 */
void 
jdoutput (char* fmt, ...)
{
  va_list args;
  va_start (args, fmt);
  socket_io_write (fmt, args);
  va_end (args);
}

/**
 * Get a charater from a connected user
 * Used by the lexer to get characters from input
 * @return the charater read
 * @see socket_io_getc
 */
char
jdgetc ()
{
  return socket_io_getc ();
}

/**
 * The entry point of a session thread.
 * When a new connection comes in a new session thread is spawned with this 
 * function.
 *
 * @param arg the argument passed in from pthread_create
 * @see socket_io_accept_connection
 * @see socket_io_close_connection
 */
static void*
main_session_thread (void *arg)
{
  int rtn;
  struct session_data_t *session_data = (struct session_data_t *) arg;
  sigset_t signals_to_block;
  sigfillset (&signals_to_block);
  pthread_sigmask (SIG_BLOCK, &signals_to_block, 0);
  pthread_mutex_lock (&connection_mutex);
  pthread_setspecific (session_key, (void*) session_data);
  socket_io_accept_connection ();
  pthread_cond_signal (&connection_signal);
  pthread_mutex_unlock (&connection_mutex);
  rtn = yyparse ();
  socket_io_close_connection ();
  session_data->active = 0;
  pthread_exit (0);
}

/**
 * Accept connections...forever
 * Here is where we start to go multi-threaded.
 * One thread per connections.
 *
 * @see socket_io_wait_connection
 * @see socket_io_deny_connection
 * @see main_session_thread
 */
static void
handle_connections ()
{
  pthread_cond_init (&connection_signal, 0);
  pthread_mutex_init (&connection_mutex, 0);
  pthread_mutex_init (&list_mutex, 0);
  pthread_mutex_lock (&connection_mutex);
  pthread_key_create (&session_key, 0);
  while (1)
    {
      struct session_data_t *session;
      socket_io_wait_connection ();

      /* make new session */
      session = (struct session_data_t *) malloc (sizeof (struct session_data_t));
      if (session) 
        {
          /* add new session to the frond of the list */
          pthread_mutex_lock (&list_mutex);
          session->active = 1;
          session->next = sessions;
          sessions = session;

          /* start the new thread */
          pthread_create (&session->thread, 0, main_session_thread, session);
          pthread_cond_wait (&connection_signal, &connection_mutex);

          /* release the list mutex */
          pthread_mutex_unlock (&list_mutex);
        }
      else
        {
          jdmessage (LOG_ERR, "insufficient memory");
          socket_io_deny_connection ();
        }

    }
}

/**
 * Shut everything down
 * @param rtncode the return to code for the OS
 */
static void
jdshutdown (int rtncode)
{
  jdmessage (LOG_DEBUG, "closing sessions");

  /* loop over list and close all sessions */
  pthread_mutex_lock (&list_mutex);
  while (sessions)
    {
      struct session_data_t *session = sessions;
      pthread_cancel (session->thread);
      pthread_join (session->thread, 0);
      sessions = session->next;
      free (session);
    }
  pthread_mutex_unlock (&list_mutex);

  term_player ();
  if (term_database) term_database ();
  term_socket_io ();
  plist_clear ();
  term_settings ();
  jdmessage (LOG_INFO, "shutdown complete");
  closelog ();
  if (rtncode != -1) exit (rtncode);
}

/**
 * Itterate over list of sessions.  Inactive sessions
 * are joined and then free'd.
 */
void
free_inactive_sessions ()
{
  struct session_data_t* session;
  struct session_data_t* previous;
  pthread_mutex_lock (&list_mutex);
  previous = sessions; 
  for (session = sessions; session; session = session->next)
    {
      if (!session->active)
        {
          pthread_join (session->thread, 0);
          if (session == sessions)
              sessions = session->next;
          else
              previous->next = session->next;
          previous == session;
        }
    }
  pthread_mutex_unlock (&list_mutex);
}

/**
 * Count active sessions.  
 */
int 
get_num_sessions ()
{
  int count = 0;
  struct session_data_t* session;
  free_inactive_sessions ();
  pthread_mutex_lock (&list_mutex);
  for (session = sessions; session; session = session->next)
    {
      count++;
    }
  pthread_mutex_unlock (&list_mutex);
  return count;
}

/**
 * Handle signals
 * @param sig the signal number recieved
 */
  static void 
sig_handler (int sig)
{
  static int got_term = 0;
  switch (sig)
    {
    case SIGSEGV:
      got_term++;
      if (got_term > 1)
        {
          jdmessage (LOG_CRIT, "got SIGSEGV inside signal handler");
          abort ();
        }
      jdmessage (LOG_CRIT, "caught segmentation fault");
      jdshutdown (-1);
      if (settings.want_core) 
        {
          jdmessage (LOG_CRIT, "dumping core");
          abort ();
        }
      exit (1);
      break;
    case SIGTERM:
    case SIGINT:
    case SIGFPE:
    case SIGBUS:
    case SIGPIPE:
    case SIGALRM:
    case SIGQUIT:
      got_term++;
      if (got_term > 1)
        {
          jdmessage (LOG_CRIT, "got sig %d inside handler", sig);
          abort ();
        }
      jdmessage (LOG_CRIT, "got signal %d : going down", sig);
      jdshutdown (1);
      break;
    case SIGHUP:
      jdmessage (LOG_INFO, "got SIGHUP");
      jdshutdown (-1);
      jdmessage (LOG_INFO, "---------------");
      initialise ();
      break;
    case SIGUSR1:
        {
          struct mallinfo info = mallinfo ();
          jdmessage (LOG_INFO, "got SIGUSR1 dumping status:");
          jdmessage (LOG_INFO, "total heap:  %d", info.arena);
          jdmessage (LOG_INFO, "mmaped regions:  %d", info.hblks);
          jdmessage (LOG_INFO, "alloced space :  %d", info.uordblks);
          break;
        }
    case SIGUSR2:
      jdmessage (LOG_CRIT, "got SIGUSR2");
      break;
    }
}

/**
 * Initialize all signal handlers
 */
static int 
init_signals ()
{
  int rtn = 0;
  sigset_t signals_to_block;
  struct sigaction sa;

  /* pthread signal mask */
  sigemptyset (&signals_to_block);
  sigaddset (&signals_to_block, SIGUSR2);
  pthread_sigmask (SIG_BLOCK, &signals_to_block, 0);

  memset (&sa, 0, sizeof (sa));
  sa.sa_handler = sig_handler;

  sigaction (SIGSEGV, &sa, NULL);
  sigaction (SIGINT , &sa, NULL);
  sigaction (SIGFPE , &sa, NULL);
  sigaction (SIGBUS , &sa, NULL);
  sigaction (SIGTERM, &sa, NULL);
  sigaction (SIGALRM, &sa, NULL);
  sigaction (SIGPIPE, &sa, NULL);
  sigaction (SIGQUIT, &sa, NULL);
  sigaction (SIGHUP , &sa, NULL);
  sigaction (SIGUSR1, &sa, NULL);
  sigaction (SIGUSR2, &sa, NULL);
  return rtn;
}
