/*
 * juked - a jukebox daemon
 * Copyright (C) 2000-2002 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  
 * Functions for parsing command line arguments
 *
 * $superduper: options.c,v 1.12 2002/05/02 09:56:16 samc Exp $
 */
#include "config.h"

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif /* HAVE_GETOPT_H */

#include "juked.h"
#include "playlist.h"

/**
 * Print the usage message to stdout.
 */
static void
print_usage ()
{
  printf ("Usage: %s [options]\n", PACKAGE_NAME);
  printf ("start the TCP/IP jukebox daemon.\n");
  printf ("\n");
  printf ("Options:\n");
  printf ("  -p, --port=<port>    Change the port that juked will listen on\n");
  printf ("  -d, --device=<dev>   Use the given audio device rather than the default\n");
  printf ("  -l, --playlist=<pl>  Load playlist at startup\n");
  printf ("  -v, --verbose        Increase verbosity level\n");
  printf ("  -r, --random         Start in random mode\n");
  printf ("  -c, --dumpcore       Dump core on seg fault\n");
  printf ("  -s, --syslog         Use syslog facility\n");
  printf ("  -a, --autostart      Autostart play (this also turns on random)\n");
  printf ("  -n, --nodaemon       Do not run as daemon\n");
  printf ("  -h, --help           display usage information\n");
  printf ("      --version        display version information\n");
  printf ("\n");
}

/**
 * Print version information.
 */
static void
print_version ()
{
  printf ("%s (Version %s)\n", PACKAGE, VERSION);
  printf ("\n");
  printf ("Copyright (C) 2001 Sam Clegg <samc@superduper.net>\n");
  printf ("This is free software; see the source for copying conditions.  There is NO\n");
  printf ("warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
}

/**
 * enum for all possible long options.
 */
enum 
{
  OPT_HELP,
  OPT_VERSION,
  OPT_DEVICE,
  OPT_RANDOM,
  OPT_AUTOSTART,
  OPT_SYSLOG,
  OPT_VERBOSE,
  OPT_LOAD,
  OPT_PORT,
  OPT_NODAEMON,
  OPT_DUMPCORE
};

static const char shortopts[] = "acrd:nhvp:l:";

static const struct option longopts[] = 
{
    { "help"     , no_argument,       0, OPT_HELP     },
    { "version"  , no_argument,       0, OPT_VERSION  },
    { "device"   , required_argument, 0, OPT_DEVICE   },
    { "helpcore" , no_argument,       0, OPT_DUMPCORE },
    { "random"   , no_argument,       0, OPT_RANDOM   },
    { "autostart", no_argument,       0, OPT_AUTOSTART},
    { "verbose"  , no_argument,       0, OPT_VERBOSE  },
    { "playlist" , required_argument, 0, OPT_LOAD     },
    { "port"     , required_argument, 0, OPT_PORT     },
    { "nodaemon" , no_argument,       0, OPT_NODAEMON },
    { "syslog"   , no_argument,       0, OPT_SYSLOG   },
    { 0          , no_argument,       0, 0            }
};

/**
 * Parse the command line options.
 * @return 0 on success, -1 on error
 */
int  
parse_opts (int argc, char** argv)
{
  int opt_error = 0;
  int index;
  char c;
  while ((c = getopt_long (argc, argv, shortopts, longopts, &index)) != -1)
    {
      switch (c) 
        {
        case 'p':
        case OPT_PORT:
          settings.port = (atoi (optarg));
          break;
        case 'l':
        case OPT_LOAD:
          plist_read (optarg);
          break;
        case 'a':
        case OPT_AUTOSTART:
          settings.autostart = 1;
          settings.random    = 1;
          break;
        case OPT_VERSION:
          print_version ();
          opt_error = -1;
          break;
        case 'h':
        case OPT_HELP:
          print_usage ();
          opt_error = -1;
          break;
        case 'r':
        case OPT_RANDOM:
          settings.random = 1;
          break;
        case 'v':
        case OPT_VERBOSE:
          settings.log_level++;
          break;
        case 'c':
        case OPT_DUMPCORE:
          settings.want_core = 1;
          break;
        case 's':
        case OPT_SYSLOG:
          settings.use_syslog = 1;
          break;
        case 'n':
        case OPT_NODAEMON:
          settings.daemon = 0;
          break;
        case 'd':
        case OPT_DEVICE:
          settings_setvalue ("device", optarg);
          break;
        case '?':
          printf ("Try `%s --help' for more information.\n", argv[0]);
          opt_error = -1;
          break;
        default:
          /* [FIXME] to work with short options 
             jdmessage (LOG_ERR,"option not yet handled: %s\n", longopts[index].name);
           */
          opt_error = -1;
          break;
        }
      if (opt_error == -1) 
        {
          settings.log_level = 1;
          break;
        }
    }

  return opt_error;
}
