/*
 * juked - a jukebox daemon
 * Copyright (C) 2000-2002 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  
 * Functions for controling mpg playback
 *
 * $superduper: player.c,v 1.25 2002/05/02 10:40:29 samc Exp $
 */
#include "config.h"

#include <signal.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <time.h>

#ifdef HAVE_SYS_SOUNDCARD_H
#include <sys/soundcard.h>
#endif /* HAVE_SYS_SOUNDCARD_H */

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */

#include "juked.h"
#include "player.h"
#include "playlist.h"
#include "commands.h"
#include "database.h"

/*
 * Include all the codec headers.
 */
#include "decode/mpglib.h"
#include "decode/vorbis.h"
#include "decode/libmad.h"

#define HISTORY_SIZE 10

/**
 * The thread that the player runs in
 */
pthread_t player_thread;       

/**
 * Pointers to decode functions, so they can be set a run-time.
 */
static int (*decoder_init) () = 0;
static int (*decoder_term) () = 0;
static int (*decoder_decode) (const char* bufin, char* bufout, 
                              int inlen, int outlen, int *result) = 0;

/** 
 * The player control thread that waits for the player to finish
 */
pthread_t play_control_thread;

/**
 * Mutexs for player
 */
pthread_mutex_t play_mutex;
pthread_mutex_t history_mutex;

/**
 * Signal that player sends when it is done playing
 */
pthread_cond_t  play_complete_cv;
pthread_mutex_t play_complete_mutex;

/**
 * Signal that will pause/unpause the player
 */
pthread_cond_t   pause_cv;
pthread_mutex_t  pause_mutex;

/**
 * Informtion on the current state of the player
 */
static struct 
{
  /* global state */
  int id;               /* ID of the track currently being played */

  /* managed my play_thread */
  int paused;           /* flag */
  int stop;             /* flag */
  int dsp;              /* fd for sound output */
  int inputfd;          /* fd to read player data from */
  int chunk_size;       /* playback chunk size (bytes) */
  char* buffer;         /* buffer for encoded audi odata */
  char* out;            /* buffer for decoded audio data */

  /* OSS stuff - not used ATM */
  int blk_size;         
  int fragment;
} player_data;

/**
 * Circular history list 
 */
static struct 
{
  int *   id;        /* list of trackid */
  time_t* time;      /* corresponding time of play */
  int     head;      /* index into list */
  int     length;    /* length of list  */
  int     playcount; /* total number of songs played */
  int     max;       /* max size of the list */
} history;


/**
 * Setup the current thread to block ALL signals
 */
static void
block_signals ()
{
  sigset_t signals_to_block;
  sigfillset (&signals_to_block);
  pthread_sigmask (SIG_BLOCK, &signals_to_block, 0);
}

/**
 * Entry point for the player control thread
 * All it really does is wait for the play_complete_cv to be signaled
 * @param arg the argument passed to pthread_create
 */
static void*
main_control_thread (void *arg)
{
  block_signals ();
  while (1)
    {
      pthread_cond_wait (&play_complete_cv, &play_complete_mutex);
      jdmessage (LOG_DEBUG, "player : play_complete signaled");
      if (settings.autoskip) cmd_next (0);
    }
  pthread_exit (0);
}

/**
 * Initialize the player_data structure
 * Open /dev/dsp and set it up
 * @return 0 on success, -1 on failure
 */
int
init_player ()
{
  struct stat stat_info;

  /* assert some preconditions */
  assert (settings.device);
  assert (settings.decoder);
  assert (settings.mp3root);

  jdmessage (LOG_INFO, "init_player");

  /* chech player root exists */
  if (stat (settings.mp3root, &stat_info) == -1)
    {
      jdmessage (LOG_ERR, "player : could not stat mp3root: %s : %s", settings.mp3root, strerror (errno));
      return -1;
    }

  /* check mp3root is a valid directory */
  if (!S_ISDIR (stat_info.st_mode))
    {
      jdmessage (LOG_ERR, "player : mp3root is not a directory: %s", settings.mp3root);
      return -1;
    }

  /* init player mutexs */
  pthread_mutex_init (&play_mutex         , 0);
  pthread_mutex_init (&history_mutex      , 0);

  /* init pause condition variable */
  pthread_mutex_init (&pause_mutex        , 0);
  pthread_cond_init  (&pause_cv           , 0);

  /* init play complete condition variable */
  pthread_mutex_init (&play_complete_mutex, 0);
  pthread_cond_init  (&play_complete_cv   , 0);

  /* start the play control thread */
  pthread_create    (&play_control_thread, 0, main_control_thread , 0);
  
  /* init the decoder */
  if (!strncmp (settings.decoder, "mad", 2))
    {
#ifdef HAVE_LIBMAD
      decoder_init     = mad_init;
      decoder_term     = mad_term;
      decoder_decode   = mad_decode;
#else
      jdmessage (LOG_ERR, "this version of %s was not compiled with mad decoder support", PACKAGE);
      return -1;
#endif /* HAVE_LIBMAD */
    }
  else if (!strncmp (settings.decoder, "vorbis", 6))
    {
#ifdef HAVE_VORBIS_CODEC_H
      decoder_init     = vorbis_init;
      decoder_term     = vorbis_term;
      decoder_decode   = vorbis_decode;
#else
      jdmessage (LOG_ERR, "this version of %s was not compiled with vorbis decoder support", PACKAGE);
      return -1;
#endif /* HAVE_VORBIS_CODEC_H */
    }
  else if (!strncmp (settings.db_type, "built-in", 8))
    {
      decoder_init     = mpglib_init;
      decoder_term     = mpglib_term;
      decoder_decode   = mpglib_decode;
    }
  else
    {
      jdmessage (LOG_ERR, "unknown database type : %s", settings.db_type);
      return -1;
    }

  /* init the history list */
  history.playcount  = 0;
  history.length     = 0;
  history.head       = -1;
  history.max        = HISTORY_SIZE;
  history.time       = (time_t *) malloc (history.max * sizeof (time_t));
  history.id         = (int *) malloc (history.max * sizeof (int));
  if (!history.time || !history.id)
    {
      free (history.time);
      free (history.id);
      jdmessage (LOG_ERR, "player : insufficient memory");
      return -1;
    }
  memset (history.id, -1, history.max);

  jdmessage (LOG_INFO, "player init : using decoder : %s", settings.decoder);
  return 0;
}

/**
 * Stop all the treads and free everything
 */
void
term_player ()
{
  jdmessage (LOG_DEBUG, "term_player");
  pthread_cancel (play_control_thread);
  pthread_join (play_control_thread, 0);
  player_stop ();
  free (history.time);
  free (history.id);
  jdmessage (LOG_DEBUG, "done term_player");
}

/**
 * Play the next "bit" of the player.  
 * This should be called in an idle loop.  
 * To pause playback just stop calling this function :)
 * @return 0 on success, -1 on failure
 */
static int
player_play_chunk ()
{
  int len, rtn;

  if (player_data.inputfd == -1) 
    {
      jdmessage (LOG_ERR, "player : no inputfd");
      return -1;
    }

  len = read (player_data.inputfd, player_data.buffer, player_data.chunk_size);
  if (len == 0) return -1; /* finished file */
  if (len < 0) 
    {
      jdmessage (LOG_ERR, "player : player inputfd: %s", strerror(errno));
      return -1; 
    }

  rtn = decoder_decode (player_data.buffer, player_data.out, 
                        len, player_data.chunk_size, &len);

  while (!player_data.stop && !player_data.paused)
    {
      int wrote_bytes = write (player_data.dsp, player_data.out, len);
      if (wrote_bytes < len)
        {
          jdmessage (LOG_ERR, "player : EEEKKKK: only wrote: %d", wrote_bytes);
        }
      if (rtn == -1) break;
      rtn = decoder_decode (0, player_data.out, 0, player_data.chunk_size, &len);
    }
  return 0;
}

#define NFRAGS 32
/**
 * Play till the EOF or stop is signalled.
 * If pause is set at any time then wait for an unpause signal 
 * @param arg the argument passed to pthread_create
 */
static void*
main_play_thread (void* arg)
{
  /* sound settings */
  int fmt     = AFMT_S16_LE;
  int stereo  = 1;
  int rate    = 44100;

  block_signals ();
  decoder_init ();

  pthread_mutex_lock (&play_mutex);
  pthread_mutex_lock (&pause_mutex);

  /* open the dsp device */
  player_data.dsp = open (settings.device, O_WRONLY);
  if (player_data.dsp < 0)
    {
      jdmessage (LOG_CRIT, "player : %s: %s", settings.device, strerror (errno));
      goto done_play_thread_nodsp;
    }
  ioctl (player_data.dsp, SNDCTL_DSP_RESET  , 0);
  ioctl (player_data.dsp, SNDCTL_DSP_SETFMT , &fmt);
  ioctl (player_data.dsp, SNDCTL_DSP_SPEED  , &rate);
  ioctl (player_data.dsp, SNDCTL_DSP_STEREO , &stereo);

  /* Allocate new buffers */
  free (player_data.buffer);
  free (player_data.out);
  player_data.buffer = (char*) malloc (player_data.chunk_size * sizeof (char*));
  player_data.out    = (char*) malloc (player_data.chunk_size * sizeof (char*));
  if (!player_data.buffer || !player_data.out)
    {
      jdmessage (LOG_ERR, "player : insufficent memory");
      goto done_play_thread;
    }

  while (!player_data.stop)
    {
      if (player_play_chunk ()) 
        {
          pthread_cond_signal (&play_complete_cv);
          break;
        }
      if (player_data.paused) 
        {
          ioctl (player_data.dsp, SNDCTL_DSP_RESET, 0);
          pthread_cond_wait (&pause_cv, &pause_mutex);
        }
    }

done_play_thread:
  /* shutdown the dsp */
  ioctl (player_data.dsp, SNDCTL_DSP_RESET, 0);
  close (player_data.dsp);
done_play_thread_nodsp:
  player_data.id      = -1; 
  close (player_data.inputfd);
  pthread_mutex_unlock (&play_mutex);
  pthread_mutex_unlock (&pause_mutex);
  decoder_term ();
  jdmessage (LOG_DEBUG, "player : player_thread finished");
  pthread_exit (0);
}

/**
 * Get the ID of the currently playing track.
 * @return the ID of the currently playing track, -1 if none
 */
int
player_get_current_id ()
{
  return player_data.id;
}

/**
 * Print out cirvular history (more recent frist)
 * TODO XMLise
 */
int
player_show_history ()
{
  int i;
  int countdown = history.length;
  if (history.head == -1) return 0;
  for (i=history.head; i>=0; i--)
    {
      jdoutput ("[%5.5d] at %s",history.id[i],ctime (&history.time[i]));
      countdown--;
    }
  if (countdown > 0)
      for (i=history.max-1; i > history.head; i--)
        {
          jdoutput ("[%5.5d] at %s\n", history.id[i], ctime (&history.time[i]));
          if (--countdown <= 0) break;
        }
  return 0;
}

/**
 * Add an element to the history
 * @param id the ID of the track to be added
 * @return 0 on success, -1 on failure
 */
int
player_history_add (int id)
{
  FILE* hist_fd;

  pthread_mutex_lock (&history_mutex);

  /* some end condition for the circular list */
  if (history.head == -1) 
      history.head = 0;
  else
      if (++history.head >= history.max) history.head = 0;

  /* now we insert at history.head */
  history.id  [history.head] = id;
  history.time[history.head] = time (&history.time[history.head]);
  if (++history.length > history.max) history.length = history.max;
  history.playcount++;

  /* write to the history file */
  if (settings.log_history)
    {
      hist_fd = fopen (settings.history_file,"a+");
      if (!hist_fd)
        {
          jdmessage (LOG_ERR,"player : failed to open history file : %s : %s",
              settings.history_file, strerror (errno));
        }
      else
        {
          fprintf (hist_fd,"[%5.5d] at %s", history.id[history.head], ctime (&history.time[history.head]));
          fflush (hist_fd);
          fclose (hist_fd);
        }
    }

  pthread_mutex_unlock (&history_mutex);

  return 0;
}

/**
 *
 */
int
player_get_playcount ()
{
  return history.playcount;
}

/**
 * Start the player thread.
 * @param id the ID of the track to be played
 * @return 0 on success, -1 on failure
 */
int 
player_start (int id)
{
  int pos;
  char filename[_POSIX_PATH_MAX];

  strncpy (filename, settings.mp3root, _POSIX_PATH_MAX-1);
  pos = strlen (filename);
  if (db_get_filename (id, filename+pos) == -1)
    {
      jdmessage (LOG_ERR,"player : database lookup filename (%d)",id);
      return -1;
    }
  player_stop ();
  jdmessage (LOG_INFO, "player : playing : [%d]", id);
  player_history_add (id);
  player_data.id         = id;
  player_data.paused     = 0;
  player_data.stop       = 0;
  player_data.chunk_size = 16 * 1024;

  player_data.inputfd = open (filename, O_RDONLY);
  if (player_data.inputfd == -1) 
    {
      jdmessage (LOG_ERR, "player : %s: %s", filename, strerror(errno));
      return -1;
    }
  if (pthread_create (&player_thread, 0, main_play_thread, 0))
    {
      jdmessage (LOG_ERR, "player : failed to create player thread");
      close (player_data.inputfd);
      return -1;
    }
  return 0;
}

/**
 * Pause playback
 * @see pause_cv
 * @return new pause state (1 = paused, 0 = unpaused), -1 on error)
 */
int 
player_pause ()
{
  if (player_data.id == -1) /* are we playing */
    {
      return -1;
    }
  else
    {
      player_data.paused = !player_data.paused;
      if (player_data.paused == 0)
        {
          pthread_cond_signal (&pause_cv);
          jdmessage (LOG_DEBUG, "player : unpausing");
        }
      else
        {
          jdmessage (LOG_DEBUG, "player : pausing");
        }
    }
  return player_data.paused;
}

/**
 * Stop playback
 * @see play_mutex
 * @see player_thread
 */
void 
player_stop ()
{
  jdmessage (LOG_DEBUG, "player : stopping");
  player_data.stop = 1;
  pthread_mutex_lock (&play_mutex);
  pthread_join (player_thread, 0);
  pthread_mutex_unlock (&play_mutex);
}
