/*
 * juked - a jukebox daemon
 * Copyright (C) 2000-2002 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  
 * Functions for manipulating playlists (linked lists).
 *
 * $superduper: playlist.c,v 1.34 2002/05/01 11:40:42 samc Exp $
 */
#include "config.h"

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */

#include "juked.h"
#include "playlist.h"
#include "commands.h"
#include "database.h"

#define DEV_RANDOM "/dev/urandom"

typedef struct listnode* nodeptr;

/**
 * A playlist is a doubly linked list of these nodes.
 */
struct listnode
{
  int track_id;
  nodeptr next;
  nodeptr prev;
};

/**
 * The list singleton.
 */
static struct 
{
  int length;
  nodeptr head;
  nodeptr current;
} thelist;

/**
 * Init the list struct.
 * @return 0 on success, -1 on failure.
 * @see struct thelist
 * @see struct listnode
 */
int
init_plist ()
{
  unsigned int seed;
  int urandom;
  jdmessage (LOG_INFO, "creating playlist");
  if ((urandom = open (DEV_RANDOM, O_RDONLY)) == -1)
    {
      jdmessage (LOG_CRIT, "failed to open %s : %s", DEV_RANDOM, strerror (errno));
      return -1;
    }
  read (urandom, (char *) &seed, 4);
  close (urandom); 
  srandom (seed);
  thelist.length   = 0;
  thelist.head     = 0;
  thelist.current  = 0;
  return 0;
}

/**
 * Insert a node into the list.
 * @param track_id the ID of the track to insert.
 * @return 0 on success, -1 on failure.
 * @see struct thelist
 * @see struct listnode
 */
int 
plist_insert (int track_id)
{
  nodeptr node = thelist.head;
  nodeptr new_node = (nodeptr)malloc (sizeof (struct listnode)); 
  new_node->track_id = track_id;
  new_node->next     = 0;
  new_node->prev     = 0;

  if (!node) thelist.head = new_node;
  else
    {
      while (node && node->next) node = node->next;
      if (node) node->next = new_node;
      new_node->prev = node;
    }
  thelist.length++;
  return 0;
}

/**
 * Remove a node from the list
 * Removes only the first instance of a track ID.
 * @param track_id the ID of the track to be removed.
 * @return 0 on success, -1 on failure.
 * @see struct thelist
 * @see struct listnode
 */   
int
plist_remove (int track_id)
{
  int rtn = -1;
  nodeptr node = thelist.head;
  while (node) 
    {
      if (node->track_id == track_id)
	{
	  if (node == thelist.head) thelist.head = node->next;
	  if (node->prev) node->prev->next = node->next;
	  if (node->next) node->next->prev = node->prev;
	  free (node);
          thelist.length--;
	  rtn = 0;
	  break;
	}
      node = node->next;
    }
  return rtn;
}

/**
 * Generate an XML version of the playlist.
 * This list MUST be free-ed somewhere else.
 * @return a pointer to a valid XML document, 0 on failure.
 */
xmlDocPtr
plist_genxml ()
{
  xmlDocPtr  doc     = 0;
  xmlNodePtr xmlnode = 0;
  char string[64];
  nodeptr node;

  doc = xmlNewDoc ("1.0");
  doc->children = xmlNewDocNode (doc, 0, XML_BASENODE_LIST, 0);

  node = thelist.head;
  while (node)
    {
      xmlnode = xmlNewChild (xmlDocGetRootElement (doc), 0, XML_TAG_TRACKID, 0);
      sprintf (string, "%d", node->track_id);
      xmlAddChild (xmlnode, xmlNewText (string));
      node = node->next;
    }

  return doc;
}

/**
 * Get the song ID of an element of the list
 * @param index the index of the element in the list (0 to length-1)
 * @return track ID at given index, 0 on failure.
 */
int 
plist_get_element_id (int index)
{
  int i;
  int rtn = -1;
  nodeptr node = thelist.head;
  for (i=0; i<index; i++)
    {
      if (node && node->next) node = node->next;
      else break;
    }
  if (i == index && node) rtn = node->track_id;
  return rtn;
}

/**
 * Advance the playlist forward.
 * @return ID of new current track in playlist.
 */
int 
plist_next ()
{
  int i;
  int rnd;
  int rtn = -1;
  if (settings.random)
    {
      if (thelist.length > 0) /* pick a random track from the list */
	{
	  rnd = (float)random ()/(float)(RAND_MAX+1.0) * (float)thelist.length;
	  thelist.current = thelist.head;
	  for (i=0;i<rnd;i++) thelist.current = thelist.current->next;
	  rtn = thelist.current->track_id;
	}
      else                    /* pick a random track from entire database */
	{
	  rtn = db_get_random_id ();
	}
    }
  else
    {
      if (thelist.length > 0)
	{
	  if (!thelist.current || !thelist.current->next) 
	      thelist.current = thelist.head;
	  else 
	      thelist.current = thelist.current->next;
          rtn = thelist.current->track_id;
	  jdmessage (LOG_DEBUG,"next track = %d\n",rtn);
	}
    }
  return rtn;
}

/**
 * Move the playlist back one track.
 * @return ID of new current track in playlist.
 */
int 
plist_prev ()
{
  int i;
  int rnd;
  int rtn = -1;

  if (settings.random)
    {
      if (thelist.length > 0) /* pick a random track from the list */
	{
	  rnd = (float)random ()/(float)RAND_MAX * (float)thelist.length;
	  thelist.current = thelist.head;
	  for (i=0;i<rnd;i++) thelist.current = thelist.current->next;
	  rtn = thelist.current->track_id;
	  jdmessage (LOG_DEBUG,"prev (random) track = %d\n",rtn);
	}
      else                    /* pick a random track from entire database */
	{
	  rtn = db_get_random_id ();
	}
    }
  else
    {
      if (thelist.length > 0)
	{
	  if (!thelist.current) 
	      thelist.current = thelist.head;
	  else if (!thelist.current->prev)
	      while (thelist.current->next) 
		  thelist.current = thelist.current->next;
	  else 
	      thelist.current = thelist.current->prev;
          rtn = thelist.current->track_id;
	  jdmessage (LOG_DEBUG,"prev track = %d\n",rtn);
	}
    }
  return rtn;
}

/**
 * Read in a playlist from a file.
 * @param filename the filename to read the playlist from.
 * @return 0 on success, -1 on failure.
 */
int 
plist_read (char* filename)
{
  xmlDocPtr  doc   = 0;
  xmlNodePtr node  = 0;
  char* string;

  doc = xmlParseFile (filename);
  if (!doc)
    {
      jdmessage (LOG_ERR,"error parsing playlist: %s\n",filename);
      return -1;
    }
  node = xmlDocGetRootElement (doc);
  if (strcmp (node->name, XML_BASENODE_LIST))
    {
      jdmessage (LOG_ERR,"config file must contain <%s> root node\n",XML_BASENODE_LIST);
      xmlFreeDoc (doc);
      return -1;
    }
  node = node->children;
  while (node != NULL)
    {
      string = xmlNodeListGetString (doc, node->xmlChildrenNode, 1);
      if (strcmp (node->name, XML_TAG_TRACKID))
        {
          jdmessage (LOG_ERR,"playlist must contain only <%s>\n",XML_TAG_TRACKID);
        }
      else
        {
          plist_insert (atoi (string));
          node = node->next;
        }
    }
  xmlFreeDoc (doc);
  return 0;
}

/**
 * Dump the current playlist to a file.
 * @param the filename to write the playlist to.
 * @return 0 on success, -1 on failure.
 */
int 
plist_dump (char* filename)
{
  xmlDocPtr doc;
  xmlChar* data;
  int size;
  int outfile = open (filename, O_WRONLY);
  if (outfile == -1)
    {
      jdmessage (LOG_ERR, "error writing playlist: %s\n", strerror (errno));
      return -1;
    }
  doc = plist_genxml ();
  xmlDocDumpFormatMemory (doc, &data, &size, 1);
  xmlFreeDoc (doc);
  if (write (outfile, data, size) == -1)
    {
      jdmessage (LOG_ERR, "error writing playlist: %s\n", strerror (errno));
      free (data);
      return -1;
    }
  free (data);
  return 0;
}

/**
 * Remove and free () all the nodes in the list.
 */
void 
plist_clear ()
{
  nodeptr node = thelist.head;
  nodeptr next;
  while (node)
    {
      next = node->next;
      free (node);
      node = next;
      thelist.length--;
    }
  if (thelist.length != 0)
    {
      jdmessage (LOG_DEBUG,"platlist.c: oops :length\n");
      thelist.length = 0;
    }
}
