/*
 * juked - a jukebox daemon
 * Copyright (C) 2000-2002 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  
 * This file contains function for parsing the configuration
 * files used by juked.  As is the convetion both system and
 * user settings are scanned.  Setting in the user file override
 * those in the system file unless "no-override" is set in the 
 * system file
 *
 * $superduper: settings.c,v 1.25 2002/05/01 11:38:08 samc Exp $
 */
#include "config.h"

#ifdef HAVE_SYS_TYPE_H
#include <sys/types.h>
#endif /* HAVE_SYS_TYPE_H */
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */

#include "juked.h"

#define RCFILE_SYSTEM "/etc/juked.conf"
#define RCFILE_USER   "/.jukedrc"

typedef enum 
{
  SETTING_STRING,
  SETTING_INT,
  SETTING_BOOL
} setting_type;

typedef struct 
{
  const char*   name;
  const char*   default_val;
  int           write_back;
  setting_type  type;
  const void*   location;
} settings_elt;

settings_t settings;

/** 
 * The internal information stored about the settings available.
 */
static settings_elt settings_info[] = 
{
  /*
   *| config option    | default value    | write? | type          | where we store it |
   *------------------------------------------------------------------------------------
   */
    { "device"         , "/dev/dsp"       , 0, SETTING_STRING, &settings.device          },
    { "historyfile"    , "./juked.history", 0, SETTING_STRING, &settings.history_file    },
    { "pwdfile"        , "./jdaccess"     , 0, SETTING_STRING, &settings.pwdfile         },
    { "mp3root"        , "/usr/share/mp3/", 0, SETTING_STRING, &settings.mp3root         },
#ifdef HAVE_MYSQL
    { "dbtype"         , "mysql"          , 0, SETTING_STRING, &settings.db_type         },
#else
    { "dbtype"         , "built-in"       , 0, SETTING_STRING, &settings.db_type         },
#endif /* HAVE_MYSQL */
    { "dbserver"       , 0                , 0, SETTING_STRING, &settings.db_server       },
    { "dbname"         , 0                , 0, SETTING_STRING, &settings.db_name         },
    { "dbuser"         , 0                , 0, SETTING_STRING, &settings.db_user         },
#ifdef HAVE_LIBMAD
    { "decoder"        , "mad"            , 0, SETTING_STRING, &settings.decoder         },
#else
    { "decoder"        , "mpglib"         , 0, SETTING_STRING, &settings.decoder         },
#endif /* HAVE_LIBMAD */
    { "rcfile"         , 0                , 0, SETTING_STRING, &settings.rcfile          },
    { "usepam"         , 0                , 0, SETTING_STRING, &settings.usepam          },
    { "dbpass"         , 0                , 0, SETTING_STRING, &settings.db_pass         },
    { "maxconnections" , "10"             , 0, SETTING_INT   , &settings.max_connections },
    { "port"           , "5853"           , 0, SETTING_INT   , &settings.port            },
    { "daemon"         , "1"              , 0, SETTING_BOOL  , &settings.daemon          },
    { "backlog"        , "10"             , 0, SETTING_INT   , &settings.backlog         },
    { "random"         , "1"              , 0, SETTING_BOOL  , &settings.random          },
    { "autoskip"       , "1"              , 0, SETTING_BOOL  , &settings.autoskip        },
    { "autostart"      , "0"              , 0, SETTING_BOOL  , &settings.autoskip        },
    { "loghistory"     , "1"              , 0, SETTING_BOOL  , &settings.log_history     },
    { "usesyslog"      , "1"              , 0, SETTING_BOOL  , &settings.use_syslog      },
    { "wantcore"       , "0"              , 0, SETTING_BOOL  , &settings.want_core       },
    { "loglevel"       , "6"              , 0, SETTING_INT   , &settings.log_level       },
    { "human"          , "0"              , 0, SETTING_INT   , &settings.human           }
};

/**
 * The number of setting available.
 */
int num_settings = sizeof (settings_info) / sizeof (settings_elt);

/**
 * Set the value of a setting. 
 * Helper function to handle malloc/free.
 * @param name the string name of the setting.
 * @param the new value of the setting in string form.
 * @return 0 on success, -1 on failure.
 */
int 
settings_setvalue (const char* name, const char* new_value)
{
  int i;
  int int_value;
  char* end;
  for (i=0; i<num_settings; i++)
    {
      if (!strcmp (name, settings_info[i].name))
        {
          switch (settings_info[i].type)
            {
            case SETTING_STRING:
              if (new_value)
                  *(char**)settings_info[i].location = strdup (new_value);
              else
                {
                  free (*(char**)settings_info[i].location);
                  *(char**)settings_info[i].location = 0;
                }
              break;
            case SETTING_INT:
              int_value = (int)strtol (&new_value[0], &end, 0);
              if (*end == '\0')
                *(int*)settings_info[i].location = int_value;
              else
                {
                  jdmessage (LOG_ERR, "invalid '%s': %s", name, new_value);
                  return -1;
                }
              break;
            case SETTING_BOOL:
              if (!strcmp (new_value,"1") || !strcmp (new_value,"true"))
                {
                  *(int*)settings_info[i].location = 1;
                }
              else if (!strcmp (new_value,"0") || !strcmp (new_value,"false"))
                {
                  *(int*)settings_info[i].location = 0;
                }
              else
                {
                  jdmessage (LOG_ERR, "invalid '%s': %s", name, new_value);
                  return -1;
                }
              break;
            default:
              jdmessage (LOG_ERR, "settings_setvalue: bad setting type (%s)", settings_info[i].name);
              break;
            }
            settings_info[i].write_back = 1;
          return 0;
        }
    }
  jdmessage (LOG_ERR, "unknown config option (%s)", name);
  return -1;
}

/**
 * Read and parse XML rcfile.
 * @param fd and open file descriptor from which to read the XML.
 * @return 0 on success, -1 on failure.
 */
static int 
settings_read (int fd)
{
  char* buffer     = 0;
  char* string     = 0;
  int   buffer_len = 0;
  int   count      = 0;
  xmlDocPtr  doc   = 0;
  xmlNodePtr node  = 0;

  buffer_len = lseek (fd, 0, SEEK_END);
  lseek (fd, 0, SEEK_SET);
  buffer = (char*)malloc ((buffer_len)*sizeof (char));
  count = read (fd, buffer, buffer_len);
  if (count != buffer_len) return 1;

  /* do XML parsing */
  doc = xmlParseMemory (buffer, buffer_len);
  if (!doc)
    {
      jdmessage (LOG_ERR, "error parsing config file");
      free (buffer);
      return -1;
    }
  node = xmlDocGetRootElement (doc);
  if (strcmp (node->name, XML_BASENODE_CONF)) 
    {
      jdmessage (LOG_ERR, "config file must contain <%s> root node", XML_BASENODE_CONF);
      xmlFreeDoc (doc);
      free (buffer);
      return -1;
    }
  node = node->children;
  while (node != NULL) 
    {
      if (!xmlIsBlankNode (node))
        {
          string = xmlNodeListGetString (doc, node->xmlChildrenNode, 1);
          if (string)
            {
              settings_setvalue (node->name, string);
              free (string);
            }
        }
      node = node->next;
    }
  xmlFreeDoc (doc);
  free (buffer);
  return 0;
}

/**
 * Write an XML rcfile based on the current settings in memory.
 * @return 0 on success, -1 on failure.
 */
int
settings_write_file (int fd)
{
  int i;
  int size;
  xmlChar*   data;
  xmlDocPtr  doc  = 0;
  xmlNodePtr node = 0;
  char string[64];
  doc = xmlNewDoc ("1.0");
  doc->children = xmlNewDocNode (doc, 0, XML_BASENODE_CONF, 0);
  for (i = 0; i < num_settings; i++)
    {
      if (settings_info[i].write_back)
        {
          node = xmlNewChild (xmlDocGetRootElement (doc), 0, settings_info[i].name, 0);
          switch (settings_info[i].type)
            {
            case SETTING_STRING:
              xmlAddChild (node, xmlNewText (*(char**) settings_info[i].location));
              break;
            case SETTING_INT:
            case SETTING_BOOL:
              snprintf (string, 64, "%d", *(int*) settings_info[i].location);
              xmlAddChild (node, xmlNewText (string));
              break;
            }
        }
    }
  xmlDocDumpFormatMemory (doc, &data, &size, 1);
  if (write (fd, data, size) == -1)
    {
      jdmessage (LOG_ERR, "error writing playlist: %s", strerror (errno));
      free (data);
      return -1;
    }
  xmlFreeDoc (doc);
  return 0;
}

/**
 * Write an XML rcfile based on the current settings in memory.
 * @return 0 on success, -1 on failure.
 */
int 
settings_write ()
{
  int out_file = open (settings.rcfile, O_WRONLY|O_CREAT);
  if (out_file == -1)
    {
      jdmessage (LOG_ERR, "failed to open rcfile for writing: %s", strerror (errno));
      return -1;
    }
  return settings_write_file (out_file);
}

/**
 * Set all settings to their defaults values.
 */
static void
settings_set_defaults ()
{
  int i;
  for (i=0; i<num_settings; i++)
    {
      if (settings_info[i].type == SETTING_STRING)
        {
          char* str_setting = * (char**) settings_info[i].location;
          /* FIXME
          if (str_setting) 
            {
              printf ("free\n");
              printf ("%s\n", str_setting);
              free (str_setting);
              printf ("done\n");
            }
            */
        }
    }
  memset (&settings, 0, sizeof (settings_t));
  for (i=0; i<num_settings; i++)
    {
      settings_setvalue (settings_info[i].name, settings_info[i].default_val);
      /* we don't want setting defaults to trigger write_back of a setting */
      settings_info[i].write_back = 0;
    }
}

/**
 * Parse the rcfiles (both user and system).
 * @return 0 on success, -1 on failure.
 */
int 
init_settings ()
{
  int rtn = 0;
  int fd  = 0;
  int found_userrc = 0;
  int found_sysrc  = 0;
  char* user_rcfile = 0;
  char* home = getenv ("HOME");

  settings_set_defaults ();

  if (home) 
    {
      user_rcfile = (char*) malloc ((strlen (home) + strlen (RCFILE_USER) + 1) * sizeof (char));
      strcpy (user_rcfile, home);
      strcat (user_rcfile, RCFILE_USER);
    }
  else
    {
      jdmessage (LOG_WARNING, "$HOME is not set");
    }

  /* read the system rcfile */
  fd = open (RCFILE_SYSTEM, O_RDONLY);
  if (fd >= 0)
    {
      found_sysrc = 1;
      jdmessage (LOG_DEBUG, "reading settings from: %s", RCFILE_SYSTEM);
      rtn |= settings_read (fd);
      close (fd);
    }
  else if (errno != ENOENT)
    {
      jdmessage (LOG_ERR, "failed to open: %s", RCFILE_SYSTEM);
    }

  /* read the user rcfile */
  if (home)
    {
      fd = open (user_rcfile, O_RDONLY);
      if (fd >= 0)
        {
          found_userrc = 1;
          jdmessage (LOG_DEBUG, "reading settings from: %s", user_rcfile);
          rtn |= settings_read (fd);
          close (fd);
        }
      else if (errno != ENOENT)
        {
          jdmessage (LOG_ERR, "failed to open: %s", RCFILE_SYSTEM);
        }
      settings_setvalue ("rcfile", user_rcfile);
    }

  /* if we found niether config file then fail */
  if (!(found_userrc || found_sysrc))
    {
      jdmessage (LOG_ERR, "could not find %s or %s", RCFILE_SYSTEM, user_rcfile);
    }
  free (user_rcfile);
  return rtn;
}

/**
 * Free all settings resourses
 */
void
term_settings ()
{
  int i;
  jdmessage (LOG_DEBUG, "term_settings");
  for (i=0; i<num_settings; i++)
    {
      if (settings_info[i].type == SETTING_STRING)
        {
          free (* (char**) settings_info[i].location);
        }
    }
}
