/*
 * juked - a jukebox daemon
 * Copyright (C) 2000-2002 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  
 * Functions for communicating with TCP/IP sockets.
 *
 * $superduper: socket_io.c,v 1.39 2002/05/02 10:39:51 samc Exp $
 */
#include "config.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "juked.h"
#include "socket_io.h"
#include "session.h"

/**
 * The file desciptor of the locally open socket.
 * The one we are listen ()ing on.
 */
static int   socketfd;
char         localhost[_POSIX_PATH_MAX];

static int addr_size = sizeof (struct sockaddr_in);

/**
 * Create, bind, and listen on the socket.
 * Allow up to settings.backlog waiting connections.
 * @return 0 on success, -1 on failure.
 */
int
init_socket_io ()
{
  int rtn = 0;
  int one = 1;
  struct sockaddr_in my_addr;

  my_addr.sin_family = AF_INET;    
  my_addr.sin_port = htons (settings.port); 
  my_addr.sin_addr.s_addr = INADDR_ANY;
  bzero (&(my_addr.sin_zero), 8); 

  socketfd = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (socketfd == -1)
    {
      jdmessage (LOG_ERR, "failed to open socket: %s", strerror (errno));
      return -1;
    }

  rtn |= setsockopt (socketfd, SOL_SOCKET, SO_REUSEADDR, (char *) &one, sizeof (one));
  rtn |= bind (socketfd, (struct sockaddr *)&my_addr, addr_size);
  rtn |= listen (socketfd, settings.backlog);
  if (rtn)
    {
      jdmessage (LOG_ERR, "failed to initializing socket");
      return -1;
    }
  if (gethostname (localhost, _POSIX_PATH_MAX) == -1)
    {
      jdmessage (LOG_ERR, "failed to lookup local hostname: %s", strerror (errno));
      return -1;
    }
  jdmessage (LOG_INFO, "listening on '%s:%d'", localhost, settings.port);
  return 0;
}

/**
 * Shutdown socket_io.  Free all resourses.
 */
void
term_socket_io ()
{
  jdmessage (LOG_DEBUG, "term_socket_io");
  if (socketfd != -1) 
    {
      close (socketfd);
      socketfd = -1;
    }
}

/**
 * Accept an incoming connection.
 * @return 0 on success, -1 on failure.
 */
int
socket_io_accept_connection ()
{
  char * ip_string;
  struct sockaddr_in addr;
  struct hostent* host_info;
  struct session_data_t* session_data = pthread_getspecific (session_key);

  session_data->remotefd = -1;
  session_data->remote_name = 0;
  session_data->outfile  = 0;

  session_data->remotefd = accept (socketfd, 0,0);
  if (session_data->remotefd == -1)
    {
      jdmessage (LOG_ERR, "socket_io : accept failed : %s", strerror (errno));
      return -1;
    }
  session_data->outfile = fdopen (session_data->remotefd, "w");
  getpeername (session_data->remotefd, (struct sockaddr *)&addr, &addr_size);
  ip_string = inet_ntoa (addr.sin_addr);
  host_info = gethostbyaddr ((char*)&addr.sin_addr, sizeof (struct in_addr), AF_INET);
  session_data->remote_name = host_info->h_name;
  jdmessage (LOG_INFO, "socket_io : client connected : %s (%s)", 
             session_data->remote_name, ip_string);
  jdoutput ("%sJuked v%s running on \"%s\". Server ready\n", OK_STR, VERSION, localhost);
  return 0;
}

/**
 * Briefly accept but send the server busy error then close connection.
 * Don't bother with error checking here we just want to close the connection.
 * @return 0 on success, -1 on failure.
 */
int
socket_io_deny_connection ()
{
  char out[256];
  int newfd;
  struct sockaddr_in addr;
  newfd = accept (socketfd, (struct sockaddr *)&addr, &addr_size);
  sprintf (out,"%sServer Busy\n",ERROR_STR);
  jdmessage (LOG_INFO, "socket_io : client denied, server busy (%s)", inet_ntoa (addr.sin_addr));
  send (newfd, out, strlen (out), 0);
  close (newfd);
  return 0;
}

/**
 * Wait for an incomming connection.
 * @return 0 on success, -1 on failure.
 */
int
socket_io_wait_connection ()
{
  int n;
  fd_set read_set;
  FD_ZERO(&read_set);
  n = socketfd + 1;
  FD_SET(socketfd, &read_set);
  while (select (n, &read_set, 0, 0, 0) == -1)
    {
      if (errno != EINTR)
        {
          jdmessage (LOG_ERR, "socket_io : select failed: %s",sys_errlist[errno]);
          return -1;
        }
    }
  return 0;
}

/**
 * Close connection to remote machine.  
 * Treadsafe.
 */
void
socket_io_close_connection ()
{
  struct session_data_t* session_data = pthread_getspecific (session_key);
  if (session_data->outfile) 
    {
      fclose (session_data->outfile);
      session_data->outfile = 0;
    }
  if (session_data->remotefd != -1) 
    {
      close (session_data->remotefd);
      session_data->remotefd = -1;
    }
}

/**
 * Write a message to the user on the other end of the socket.
 * Threadsafe.
 * @param fmt the format of the message (printf)
 * @param the positional paramaters for the message.
 */
void
socket_io_write (char* fmt, va_list args)
{
  struct session_data_t* session_data = pthread_getspecific (session_key);
  if (!session_data || !session_data->outfile) return;
  vfprintf (session_data->outfile,fmt,args);
  fflush (session_data->outfile);
}

/**
 * Get a character from the other end of the connection.  
 * Can block.
 * Threadsafe.
 * @return the charater recieved.
 */
char
socket_io_getc ()
{ 
  int rtn;
  char c;
  struct session_data_t* session_data = pthread_getspecific (session_key);

  while (1)
    {
      rtn = read (session_data->remotefd, &c, 1);
      if (rtn == -1 || rtn == 0) 
        {
          jdmessage (LOG_INFO, "socket_io : client disconnected : %s", 
              session_data->remote_name);
          return EOF;
        }
      break;
    }
  return c;
}
